from .atomic_ import atomic
from .context import AsyncContext
from .context import Context
from .interfaces.adapter import Adapter
from .interfaces.adapter import AsyncAdapter
from .interfaces.adapter import AtomicAdapter
from .interfaces.adapter import AtomicAsyncAdapter

__all__ = [
    "Context",
    "AsyncContext",
    "AsyncAdapter",
    "Adapter",
    "atomic",
    "AtomicAsyncAdapter",
    "AtomicAdapter",
]



from typing import Protocol, TypeVar, cast

from lazyfields import lazyfield

from gyver.exc import InvalidPath

from .helpers import python_filename
from .typedef import BaseFile, File, Folder, TextFile

FileT = TypeVar("FileT", bound=BaseFile)

T = TypeVar("T")


class AbstractFileTree(Protocol[T]):
    base_dir: T

    @lazyfield
    def root(self) -> Folder:
        raise NotImplementedError

    def create_dir(self, dirname: str, *path: str) -> Folder:
        target = self.root
        for item in path:
            if item not in target.contents:
                target.contents[item] = Folder(item)
            target = target.contents[item]
            if isinstance(target, File):
                raise InvalidPath("Foldername conflicts with file")
        if dirname not in target.contents:
            target.contents[dirname] = Folder(dirname)
        if not isinstance(target.contents[dirname], Folder):
            raise InvalidPath("foldername conflicts with filename")
        return cast(Folder, target.contents[dirname])

    def create_file(
        self, filename: str, *path: str, fileclass: type[FileT] = File
    ) -> FileT:
        if path:
            dirname, *restpath = path
            directory = self.create_dir(dirname, *restpath)
        else:
            directory = self.root
        target = directory.contents.get(filename) or fileclass(filename)
        if not isinstance(target, fileclass):
            raise InvalidPath(
                "filename conflicts with other folder, or different fileclass"
            )
        directory.contents[filename] = target
        return target

    def create_text_file(self, filename: str, *path: str) -> TextFile:
        return self.create_file(filename, *path, fileclass=TextFile)

    def get_file(self, filename: str, *path: str) -> File:
        if path:
            dirname, *restpath = path
            directory = self.get_dir(dirname, *restpath)
        else:
            directory = self.root
        if not directory:
            raise InvalidPath("file not found")
        target = directory.contents.get(filename)
        if not target:
            raise InvalidPath("file not found")
        if not isinstance(target, File):
            raise InvalidPath("filename conflicts with folder")
        return target

    def get_dir(self, dirname: str, *path: str) -> Folder | None:
        target = self.root
        for item in path:
            if item not in target.contents:
                return None
            target = target.contents[item]
            if isinstance(target, File):
                raise InvalidPath("Foldername conflicts with file")
        return cast(Folder | None, target.contents.get(dirname))

    def create_py_file(
        self,
        filename: str,
        *path: str,
        private: bool = False,
        dunder: bool = False,
    ) -> TextFile:
        return self.create_file(
            python_filename(filename, private, dunder),
            *path,
            fileclass=TextFile,
        )

    def dunder_init(self, *path: str):
        return self.create_py_file("init", *path, dunder=True)



from typing import Generic

from lazyfields import dellazy
from lazyfields import lazyfield

from gyver.context.context import AsyncContext
from gyver.context.context import Context
from gyver.context.interfaces.adapter import AtomicAdapter
from gyver.context.interfaces.adapter import AtomicAsyncAdapter
from gyver.context.typedef import T


class AtomicContext(Context[T], Generic[T]):
    adapter: AtomicAdapter[T]

    def __init__(self, adapter: AtomicAdapter[T]) -> None:
        super().__init__(adapter)

    @lazyfield
    def client(self):
        return self.adapter.new()

    def acquire(self):
        with self._lock:
            if self.adapter.is_closed(self.client):
                dellazy(self, "client")
            self._stack += 1
            self.adapter.begin(self.client)
        return self.client

    def release(self, commit: bool = True):
        with self._lock:
            if self._stack == 1:
                if self.adapter.in_atomic(self.client):
                    if commit:
                        self.adapter.commit(self.client)
                    else:
                        self.adapter.rollback(self.client)
                self.adapter.release(self.client)
                dellazy(self, "client")
            self._stack -= 1

    def __exit__(self, *exc):
        self.release(not any(exc))


class AsyncAtomicContext(AsyncContext[T], Generic[T]):
    adapter: AtomicAsyncAdapter[T]

    def __init__(self, adapter: AtomicAsyncAdapter[T]) -> None:
        super().__init__(adapter)

    async def acquire(self):
        async with self._lock:
            client = await self.client()
            if self.stack == 0:
                await self.adapter.begin(client)
            self._stack += 1
            return client

    async def release(self, commit: bool = True):
        async with self._lock:
            client = await self.client()
            if self._stack == 1:
                if await self.adapter.in_atomic(client):
                    if commit:
                        await self.adapter.commit(client)
                    else:
                        await self.adapter.rollback(client)
                await self.adapter.release(client)
                dellazy(self, "client")
            self._stack -= 1

    async def __aexit__(self, *exc):
        await self.release(not any(exc))



import sys
from itertools import chain
from collections.abc import Sequence


class GyverError(Exception):
    """Base exception for all exceptions raised by the Gyver library."""


class InvalidCast(GyverError):
    """Raised when an error occurs during a casting operation in configuration."""


class MissingName(GyverError, KeyError):
    """Raised when a required name is missing in the provided environment configuration."""


class InvalidField(GyverError, KeyError):
    """Raised when an unexpected result is encountered during a field lookup on a mapping."""


class CacheMiss(GyverError, KeyError):
    """Raised when a cache lookup returns None or raises an error."""


class MissingParams(GyverError, NotImplementedError):
    """Raised when a function expecting *args parameters receives none."""


class InvalidPath(GyverError, ValueError):
    """Raised when an invalid file or directory path is encountered."""


class InvalidParamType(GyverError, TypeError):
    """Raised when a parameter's type is not as expected."""


class InvalidParamValue(GyverError, ValueError):
    """Raised when a parameter's value is unexpected."""


if sys.version_info < (3, 11):

    class ErrorGroup(GyverError):
        """
        Exception that groups multiple sub-exceptions.

        This exception is used to encapsulate and present multiple related exceptions as a group.
        It provides a way to handle multiple errors together in a unified manner.
        """

        exceptions: tuple[Exception, ...]

        def __init__(self, message: str, exceptions: Sequence[Exception]) -> None:
            self.message = message
            self.exceptions = tuple(exceptions)
            args = list(chain(item.args for item in exceptions))
            super().__init__(self.message, *args)

        def __str__(self) -> str:
            suffix = "" if len(self.exceptions) == 1 else "s"
            return f"{self.message} ({len(self.exceptions)} sub-exception{suffix})"

        def __repr__(self) -> str:
            return f"{self.__class__.__name__}({self.message!r}, {self.exceptions!r})"

else:

    class ErrorGroup(GyverError, ExceptionGroup):  # noqa: F821
        """
        Exception that groups multiple sub-exceptions.

        This exception is used to encapsulate and present multiple related exceptions as a group.
        It provides a way to handle multiple errors together in a unified manner.
        """


class FailedFileOperation(GyverError):
    """Raised when an error occurs during an interaction with a file."""


class MergeConflict(GyverError):
    """Raised when a merge conflict is detected, e.g., during the merge_dicts operation."""



import sys
from abc import ABC, abstractmethod
from io import SEEK_SET, BytesIO
from typing import Generic

import cchardet
from typing_extensions import Self, TypeVar, override

from gyver.attrs import asdict, call_init, define

T = TypeVar("T")
BoundType = TypeVar("BoundType", str, bytes)


@define
class VirtualPath(Generic[T]):
    name: str
    contents: T


class BaseFile(VirtualPath[BytesIO], Generic[BoundType], ABC):
    def __init__(self, name: str):
        super().__init__(name, BytesIO())

    def get_encoding(self) -> str:
        val = cchardet.detect(self.file_contents())
        self.seek(SEEK_SET)
        return val["encoding"] or sys.getdefaultencoding()

    @abstractmethod
    def read(self, size: int = -1) -> BoundType: ...

    @abstractmethod
    def readline(self, size: int = -1) -> BoundType: ...

    @abstractmethod
    def readlines(self, hint: int = -1) -> list[BoundType]: ...

    @abstractmethod
    def write(self, content: BoundType) -> int: ...

    @abstractmethod
    def writelines(self, lines: list[BoundType]): ...

    def seek(self, offset: int, whence: int = SEEK_SET) -> int:
        return self.contents.seek(offset, whence)

    def tell(self) -> int:
        return self.contents.tell()

    def flush(self):
        self.contents.flush()

    def close(self):
        self.contents.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *_):
        self.close()

    def file_contents(self):
        return self.contents.getvalue()

    def __parse_dict__(self, by_alias: bool):
        return {"name": self.name, "contents": self.contents}


class File(BaseFile[bytes]):
    @override
    def read(self, size: int = -1) -> bytes:
        return self.contents.read(size)

    @override
    def readline(self, size: int = -1) -> bytes:
        return self.contents.readline(size)

    @override
    def readlines(self, hint: int = -1) -> list[bytes]:
        return self.contents.readlines(hint)

    @override
    def write(self, content: bytes) -> int:
        return self.contents.write(content)

    @override
    def writelines(self, lines: list[bytes]):
        self.contents.writelines(lines)


@define
class TextFile(BaseFile[str]):
    encoding: str

    def __init__(self, name: str, encoding: str = "utf-8"):
        call_init(self, name, BytesIO(), encoding)

    @override
    def read(self, size: int = -1) -> str:
        return self.contents.read(size).decode(self.encoding)

    @override
    def readline(self, size: int = -1) -> str:
        return self.contents.readline(size).decode(self.encoding)

    @override
    def readlines(self, hint: int = -1) -> list[str]:
        return [line.decode(self.encoding) for line in self.contents.readlines(hint)]

    @override
    def write(self, content: str) -> int:
        return self.contents.write(content.encode(self.encoding))

    @override
    def writelines(self, lines: list[str]):
        encoded_lines = [line.encode(self.encoding) for line in lines]
        self.contents.writelines(encoded_lines)


class Folder(VirtualPath[dict[str, VirtualPath]]):
    def __init__(self, name: str):
        super().__init__(name, {})

    def __parse_dict__(self, by_alias: bool):
        return {
            "name": self.name,
            "contents": {
                key: asdict(value, by_alias) for key, value in self.contents.items()
            },
        }
